/*
seams.c
Author: Hayden Walles
Date: 31 August 2007
A demonstration of least-energy seam removal for content-aware image resizing.
*/

/*
    Copyright (C) 2007, 2008, 2009 Hayden Walles

    This file is part of Seamstress.

    Seamstress is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Seamstress is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "seams.h"


/*
For a linked-list representation:
 {store the pixels in an IMAGE.  They aren't needed for seam computation.  Compute the energy then copy it into the intial nodes array -OR- leave it in a separate array if that is preferable}
EITHER:
typedef struct tagNODE{
unsigned int myergs,ergs;
int nextindex;
struct tagNODE *left,*right,*up,*down;
} NODE;
OR
typedef struct tagNODE{
unsigned int ergs;
int nextindex;
struct tagNODE *left,*right,*up,*down;
}
Note that because the structures are still held in an array, the original position of any NODE can be computed via pointer arithmetic as querynode-firstnodeofrow (similar for other direction).

Of course, the other option is to do a copy after each seam to compact the array into a rectangular form.  In that case, there is the overhead of the copy, but the NODE can be reduced to
typedef struct {
unsigned int ergs,originalcoord;
int nextindex;
}
--originalcoord is now required so we can keep track of which pixel this was originally.  Furthermore, if we assume that coordinates range up to ~2047 max (ie <2^11), and we know that nextindex needs only two bits (range [-1,1]) which leaves 19 bits for ergs (and if pixel energy values are <256, that leaves space for seams 2048 pixels long - just right!) then a node can exist via bitfields as:
typedef struct {
unsigned int ergs:19;
unsigned int originalcoord:11;
int nextindex:2;
} NODE;
which is only 32 bits.  Of course similar compression can be used for the pointer version, with a resulting 20 byte structure.  It all depends on how bad the copying overhead is, I suppose.
*/


#define NODE_AT(picker,row,col) (picker->nodes[picker->common.width*(row)+(col)])




SEAM_UNPICKER *seamstressNewUnpicker(int actualwidth, int actualheight,int direction,int energyfunc,int dynamicenergy,int *error){
  SEAM_UNPICKER *new;
  int i,j,ei,ej;
  int *colmap;
  int left,right;
  int width,height;

  SET_ERROR(SEAMSTRESS_E_OK);

  if((actualwidth>SEAMSTRESS_UNPICKER_MAX_SIZE)||(actualheight>SEAMSTRESS_UNPICKER_MAX_SIZE)){
    SET_ERROR(SEAMSTRESS_E_TOOBIG);
    return NULL;
  }
  
  if(direction==0){
    width=actualwidth;
    height=actualheight;
  }
  else {
    width=actualheight;
    height=actualwidth;
  }

  if(!gaussianGradientEnergyInitialise(error))
    return NULL;
 
  
  new=malloc(sizeof(SEAM_UNPICKER)+sizeof(NODE)*height*width+sizeof(int)*height);
  
  if(new==NULL){
    SET_ERROR(SEAMSTRESS_E_MEMORY);
    return NULL;
  }
  new->nodes=(NODE *)(((char *)new)+sizeof(SEAM_UNPICKER));
  new->lastseam=(int *)(((char *)new->nodes)+sizeof(NODE)*height*width);
  new->lastfrom=new->lastto=0;
  new->seamcount=0;
  new->direction=direction;

  new->common.energy=NULL;
  new->dynamicenergy=dynamicenergy;
  new->common.ergprivate=NULL;
  new->common.marks=NULL;
  new->common.luminance=NULL;

 
  /*Initialise the node offsets*/

  new->actualwidth=actualwidth;
  new->actualheight=actualheight;
  new->common.width=width;
  new->common.height=height;
 
  
  for(i=0;i<height;i++){
    for(j=0;j<width;j++){
      new->nodes[i*width+j].offset=j;
    }
  }
    
  /*Initialise the energy function, and the luminance matrix.*/
  if(!gaussianGradientEnergyStart(&new->common,error)){
    free(new);
    return NULL;
  }
  
  return new;
  
}


void unpickerSetRowRGB24(SEAM_UNPICKER *pic, int row, unsigned char *pixels){
  int hborder,vborder;

  hborder=(matrixCols(pic->common.luminance)-pic->common.width)/2;
  vborder=(matrixRows(pic->common.luminance)-pic->common.height)/2;
 
  setPictureRowRGB24(pic->common.luminance,hborder,vborder,pic->direction,row,pixels);

  return;
}

void unpickerSetRowRGB32(SEAM_UNPICKER *pic, int row, unsigned char *pixels){
  
  int hborder,vborder;
  
  hborder=(matrixCols(pic->common.luminance)-pic->common.width)/2;
  vborder=(matrixRows(pic->common.luminance)-pic->common.height)/2;
  
  setPictureRowRGB32(pic->common.luminance,hborder,vborder,pic->direction,row,pixels);

  return;

}



void unpickerDelete(SEAM_UNPICKER *pic){
  
  gaussianGradientEnergyFinish(&pic->common);
  if(pic->common.energy)
    deleteMatrix(pic->common.energy);
  if(pic->common.luminance)
    deleteMatrix(pic->common.luminance);
    
  free(pic);

}




int minimalSeams(SEAM_UNPICKER *pic){
  NODE *currow,*oldrow;
  int i,j;
  int *curergs,*oldergs;
  int ewidth=pic->common.width;
  int min;

  oldergs=pic->common.energy->buffer.integers;
  curergs=oldergs+pic->common.width;
  oldrow=pic->nodes;
  currow=oldrow+pic->common.width;

  /*Initialise the bottom row.*/
  for(j=0;j<ewidth;j++)
    oldrow[j].ergs=oldergs[j];
  
  /*Now back the values up through the image.*/
  for(i=1;i<pic->common.height;i++){
      
    for(j=0;j<ewidth;j++){
      min=oldrow[j].ergs;
      if(j>0){
	if(oldrow[j-1].ergs<min)
	  min=oldrow[j-1].ergs;
      }
      if(j<ewidth-1){
	if(oldrow[j+1].ergs<min)
	  min=oldrow[j+1].ergs;
      }
      currow[j].ergs=min+curergs[currow[j].offset];
    }

    oldrow=currow;
    currow+=pic->common.width;
    curergs+=pic->common.width;
  }
  
  return 1;

}


/*Does the same thing as minimalVertical seams, but does it incrementally, recomputing only those regions around the most recently removed seam.*/
int partialMinimalSeams(SEAM_UNPICKER *pic){
  NODE *currow,*oldrow;
  int i,j;
  int *curergs,*oldergs;
  int ewidth=pic->common.width-pic->seamcount;
  int l,r;
  int min;

  oldergs=pic->common.energy->buffer.integers;
  curergs=oldergs+pic->common.width;
  oldrow=pic->nodes;
  currow=oldrow+pic->common.width;

  if(pic->dynamicenergy){
    /*Make sure that the energy stored in the initial row of nodes is correct where we've recomputed it it.*/
    for(j=pic->lastfrom;j<=pic->lastto;j++)
      oldrow[j].ergs=oldergs[oldrow[j].offset];
  }

  l=pic->lastfrom;//pic->lastseam[0];
  r=pic->lastto;//l+1;

 for(i=1;i<pic->common.height;i++){
  
    l=MAX(l-1,0);
    r=MIN(r+1,ewidth);
    for(j=l;j<r;j++){
      min=oldrow[j].ergs;
      if((j>0)&&(oldrow[j-1].ergs<min))
	min=oldrow[j-1].ergs;
      if((j<ewidth-1)&&(oldrow[j+1].ergs<min))
	min=oldrow[j+1].ergs;
      currow[j].ergs=min+curergs[currow[j].offset];
    }

    oldrow=currow;
    currow+=pic->common.width;
    curergs+=pic->common.width;
 }

  return 1;

}


int findMinimalSeam(SEAM_UNPICKER *pic, int *indices){
  NODE *currow;
  int i,j;
  int diff;
  int minindex;
  int ewidth=pic->common.width-pic->seamcount;
  
  /*We have now solved the problem.  Just read the seam off into the indices array.*/
  /*We assume the minimal seams through each of the bottom pixels have already been computed; we just find the minimal one of those.*/
  
  currow=pic->nodes+((pic->common.height-1)*pic->common.width);
  minindex=0;

  for(j=1;j<ewidth;j++){
    if(currow[j].ergs<currow[minindex].ergs)
      minindex=j;
  }


  indices[pic->common.height-1]=minindex;
  
   /*modify neighbours of the new seam, as necessary.*/
  pic->lastseam[pic->common.height-1]=indices[pic->common.height-1];

  for(i=pic->common.height-2;i>=0;i--){
    diff=currow[indices[i+1]].ergs-getInteger(pic->common.energy,i+1,currow[indices[i+1]].offset);

    currow-=pic->common.width;
    if((indices[i+1]>0)&&(currow[indices[i+1]-1].ergs==diff))
      indices[i]=indices[i+1]-1;
    else if((indices[i+1]<ewidth-1)&&(currow[indices[i+1]+1].ergs==diff))
      indices[i]=indices[i+1]+1;
    else 
      indices[i]=indices[i+1];

    pic->lastseam[i]=indices[i];

  }
  pic->seamcount++;

  pic->lastfrom=pic->lastto=pic->lastseam[0];
  pic->lastto++;

  return 1;

}


int removeSeam(SEAM_UNPICKER *inpic, int *seam,int translate){
  int i,j;
  NODE *rowptr;
  int idx;
  /*The translation phase here could be separated out.*/

  rowptr = inpic->nodes;
  
  for(i=0;i<inpic->common.height;i++){
    idx=seam[i];
    if(translate)
      seam[i]=rowptr[idx].offset;
    for(j=idx;j<inpic->common.width-inpic->seamcount;j++)
      rowptr[j]=rowptr[j+1];
    
    rowptr+=inpic->common.width;
  }

  return 1;

}


/*seam must have room for at least pic->common.height entries.  If translate=1 then the seam returned is in terms of the original image, otherwise the seam returned is in terms of the image with previously identified seams removed.*/
int unpickerNextSeam(SEAM_UNPICKER *pic, int *seam,int translate,int *error){
 
  if(pic->seamcount>=pic->common.width){
    SET_ERROR(SEAMSTRESS_E_FINISHED);
    return 0;
  }

  if(pic->seamcount==0)
    minimalSeams(pic);
  else
    partialMinimalSeams(pic);

  findMinimalSeam(pic,seam);
  if(pic->dynamicenergy)
    adjustEnergyVertically(pic,pic->seamcount,0,seam);
  removeSeam(pic,seam,translate);
  SET_ERROR(SEAMSTRESS_E_OK);
  return 1;
  
}

SEAM_MAP *unpickerMap(SEAM_UNPICKER *pic,void (*updater)(int done,int of,void *user),void *user, int *error){
  int i,j;
  SEAM_MAP *map;
  int n;
#ifdef HAVE_DYNAMIC_ARRAYS
  int seam[pic->common.height];
#elif defined(HAVE_ALLOCA)
  int *seam;
  seam=(int *)alloca(sizeof(int)*pic->common.height);
#else
  /*I should replace this with a heap-allocated array, of course.*/
#error Need dynamic arrays or alloca!
#endif



  n=pic->common.width;

  
  if(pic->common.energy==NULL){
    SET_ERROR(SEAMSTRESS_E_NOENERGY);
    return NULL;
  }

  map=malloc(sizeof(SEAM_MAP));
  if(map==NULL){
    SET_ERROR(SEAMSTRESS_E_MEMORY);
    return NULL;
  }
  

  map->actualwidth=pic->actualwidth;
  map->actualheight=pic->actualheight;
  map->direction=pic->direction;
  
  if((map->map=newMatrix(map->actualheight,map->actualwidth,MT_INTEGER,error))==NULL){
    SET_ERROR(SEAMSTRESS_E_MEMORY);
    free(map);
    return NULL;
  }
  
  
  for(i=0;i<n;i++){
    if(!unpickerNextSeam(pic,seam,1,error)){
      mapDelete(map);
      return NULL;
    }
    
    if(pic->direction==0){
      for(j=0;j<pic->common.height;j++)
	setInteger(map->map,j,seam[j],i+1);
    }
    else {
      for(j=0;j<pic->common.height;j++)
	setInteger(map->map,seam[j],j,i+1);
    }
    if(updater!=NULL)
      updater(i+1,n,user);
  }
  SET_ERROR(SEAMSTRESS_E_OK);
  return map;

}


void mapDelete(SEAM_MAP *map){
  deleteMatrix(map->map);
  free(map);
}

void unpickerCurrentSize(SEAM_UNPICKER *picker, int *widthptr,int *heightptr){

  if(picker->direction==0){
    *widthptr=picker->common.width-picker->seamcount;
    *heightptr=picker->common.height;
  }
  else {
    *heightptr=picker->common.width-picker->seamcount;
    *widthptr=picker->common.height;
  }
}

/*Converts (*xptr,*yptr) from the current image coordinates to the original image coordinates.*/
int unpickerTransformToOriginal(SEAM_UNPICKER *pic, int *xptr, int *yptr){
  unsigned x,y;

  x=*xptr;
  y=*yptr;

  if(pic->direction==0){
    if((y>=pic->common.height)||(x>=pic->common.width-pic->seamcount))
      return 0;
    *xptr=NODE_AT(pic,y,x).offset;
  }
  else {
    if((x>=pic->common.height)||(y>=pic->common.width-pic->seamcount))
      return 0;
    *yptr=NODE_AT(pic,x,y).offset;
  }
  
  return 1;
}
