/*
unweaver.c
Author: Hayden Walles
Date: 18 September 2007
Implementation of the unweaver interface, the more general of the two seam carving interfaces.
*/

/*
    Copyright (C) 2007 Hayden Walles

    This file is part of Seamstress.

    Seamstress is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Seamstress is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "seams.h"

#define NODE_AT(weaver,row,col) (weaver->nodes[weaver->common.width*(row)+(col)])
#define ERGS_AT(weaver,row,col) getInteger(weaver->common.energy,row,col)

SEAM_UNWEAVER *seamstressNewUnweaver(int actualwidth, int actualheight,int energyfunc,int dynamicenergy,int *error){
  SEAM_UNWEAVER *new;
  int i,j,ei,ej;
  int *colmap;
  int left,right;
  int width,height;

  *error = MATRIX_E_OK;
  
  if(dynamicenergy){
    if(error)
      *error=SEAMSTRESS_E_UNSUPPORTED;
    return NULL;
  }

  width=actualwidth;
  height=actualheight;
  
  if(!gaussianGradientEnergyInitialise(error))
    return NULL;
  
  
  new=malloc(sizeof(SEAM_UNWEAVER)+sizeof(BIGNODE)*height*width);
  
  if(new==NULL){
    return NULL;
  }
  new->nodes=(BIGNODE *)(((char *)new)+sizeof(SEAM_UNPICKER));
  new->hseamcount=new->vseamcount=0;

  new->common.energy=NULL;
  new->dynamicenergy=dynamicenergy;
  new->common.ergprivate=NULL;
  new->common.marks=NULL;
  new->common.luminance=NULL;
 
  /*Initialise the node offsets*/

  new->common.width=width;
  new->common.height=height;
 
  for(i=0;i<new->common.height;i++){
    for(j=0;j<new->common.width;j++){
      new->nodes[i*new->common.width+j].xoffset=j;
      new->nodes[i*new->common.width+j].yoffset=i;
    }
  }
    
  /*Initialise the energy function, and the luminance matrix.*/
  if(!gaussianGradientEnergyStart(&new->common,error)){
    free(new);
    return NULL;
  }
  
  return new;
  
}


void unweaverSetRowRGB24(SEAM_UNWEAVER *pic, int row, unsigned char *pixels){
  int hborder,vborder;

  hborder=(matrixCols(pic->common.luminance)-pic->common.width)/2;
  vborder=(matrixRows(pic->common.luminance)-pic->common.height)/2;
 
  setPictureRowRGB24(pic->common.luminance,hborder,vborder,0,row,pixels);

  return;
}

void unweaverSetRowRGB32(SEAM_UNWEAVER *pic, int row, unsigned char *pixels){
  
  int hborder,vborder;
  
  hborder=(matrixCols(pic->common.luminance)-pic->common.width)/2;
  vborder=(matrixRows(pic->common.luminance)-pic->common.height)/2;
  
  setPictureRowRGB32(pic->common.luminance,hborder,vborder,0,row,pixels);

  return;

}


int unweaverDelete(SEAM_UNWEAVER *weaver){
  
 gaussianGradientEnergyFinish(&weaver->common);
  if(weaver->common.energy)
    deleteMatrix(weaver->common.energy);
  if(weaver->common.luminance)
    deleteMatrix(weaver->common.luminance);
    
  free(weaver);


}




int minimalVerticalSeams(SEAM_UNWEAVER *pic){
  BIGNODE *currow,*oldrow;
  int i,j;
  int ewidth=pic->common.width-pic->vseamcount;
  int eheight=pic->common.height-pic->hseamcount;
  int min;

  oldrow=pic->nodes;
  currow=oldrow+pic->common.width;

  /*Initialise the top row.*/
  for(j=0;j<ewidth;j++)
    oldrow[j].ergs=ERGS_AT(pic,oldrow[j].yoffset,oldrow[j].xoffset);
  /*Now back the values up through the image.*/
  for(i=1;i<eheight;i++){
      
    for(j=0;j<ewidth;j++){
      min=oldrow[j].ergs;
      if(j>0){
	if(oldrow[j-1].ergs<min)
	  min=oldrow[j-1].ergs;
      }
      if(j<ewidth-1){
	if(oldrow[j+1].ergs<min)
	  min=oldrow[j+1].ergs;
      }
      currow[j].ergs=min+ERGS_AT(pic,currow[j].yoffset,currow[j].xoffset);
    }

    oldrow=currow;
    currow+=pic->common.width;
  }
  
  return 1;

}


int minimalHorizontalSeams(SEAM_UNWEAVER *pic){
  int i,j;
  int eheight=pic->common.height-pic->hseamcount;
  int ewidth=pic->common.width-pic->vseamcount;
  int min;


  /*Initialise the left row.*/
  for(i=0;i<eheight;i++)
    NODE_AT(pic,i,0).ergs=ERGS_AT(pic,NODE_AT(pic,i,0).yoffset,NODE_AT(pic,i,0).xoffset);

  /*Now back the values up through the image.*/
  for(j=1;j<ewidth;j++){
      
    for(i=0;i<eheight;i++){
      min=NODE_AT(pic,i,j-1).ergs;
      if(i>0){
	if(NODE_AT(pic,i-1,j-1).ergs<min)
	  min=NODE_AT(pic,i-1,j-1).ergs;
      }
      if(i<eheight-1){
	if(NODE_AT(pic,i+1,j-1).ergs<min)
	  min=NODE_AT(pic,i+1,j-1).ergs;
      }
      NODE_AT(pic,i,j).ergs=min+ERGS_AT(pic,NODE_AT(pic,i,j).yoffset,NODE_AT(pic,i,j).xoffset);
    }
  }
  
  return 1;

}



int findMinimalVerticalSeam(SEAM_UNWEAVER *pic, int *indices){
  int i,j;
  int diff;
  int minindex;
  int ewidth=pic->common.width-pic->vseamcount;
  int eheight=pic->common.height-pic->hseamcount;

  /*We have now solved the problem.  Just read the seam off into the indices array.*/
  /*We assume the minimal seams through each of the bottom pixels have already been computed; we just find the minimal one of those.*/
  
  minindex=0;

  for(j=1;j<ewidth;j++){
    if(NODE_AT(pic,eheight-1,j).ergs<NODE_AT(pic,eheight-1,minindex).ergs)
      minindex=j;
  }


  indices[eheight-1]=minindex;

  /*Trace the seam back up to the top, noting where each member pixel lies.*/
  for(i=eheight-2;i>=0;i--){
    diff = NODE_AT(pic,i+1,indices[i+1]).ergs - getInteger(pic->common.energy,NODE_AT(pic,i+1,indices[i+1]).yoffset,NODE_AT(pic,i+1,indices[i+1]).xoffset);
    
    if((indices[i+1]>0)&&(NODE_AT(pic,i,indices[i+1]-1).ergs==diff))
      indices[i]=indices[i+1]-1;
    else if((indices[i+1]<ewidth-1)&&(NODE_AT(pic,i,indices[i+1]+1).ergs==diff))
      indices[i]=indices[i+1]+1;
    else 
      indices[i]=indices[i+1];

  }
  pic->vseamcount++;

  return 1;

}


int findMinimalHorizontalSeam(SEAM_UNWEAVER *pic, int *indices){
  int i,j;
  int diff;
  int minindex;
  int eheight=pic->common.height-pic->hseamcount;
  int ewidth=pic->common.width-pic->vseamcount;

  /*We have now solved the problem.  Just read the seam off into the indices array.*/
  /*We assume the minimal seams through each of the rightmost pixels have already been computed; we just find the minimal one of those.*/
  
  minindex=0;

  for(i=1;i<eheight;i++){
    if(NODE_AT(pic,i,ewidth-1).ergs<NODE_AT(pic,minindex,ewidth-1).ergs)
      minindex=i;
  }


  indices[ewidth-1]=minindex;

  /*Trace the seam back over to the left, noting where each member pixel lies.*/
  for(j=ewidth-2;j>=0;j--){
    diff = NODE_AT(pic,indices[j+1],j+1).ergs - getInteger(pic->common.energy,NODE_AT(pic,indices[j+1],j+1).yoffset,NODE_AT(pic,indices[j+1],j+1).xoffset);
    
    if((indices[j+1]>0)&&(NODE_AT(pic,indices[j+1]-1,j).ergs==diff))
      indices[j]=indices[j+1]-1;
    else if((indices[j+1]<eheight-1)&&(NODE_AT(pic,indices[j+1]+1,j).ergs==diff))
      indices[j]=indices[j+1]+1;
    else 
      indices[j]=indices[j+1];

  }
  pic->hseamcount++;

  return 1;

}



int removeVerticalSeam(SEAM_UNWEAVER *inpic, int *seam){
  int i,j;
  BIGNODE *rowptr;
  int idx;
  int eheight = inpic->common.height-inpic->hseamcount;

  rowptr = inpic->nodes;
  
  for(i=0;i<eheight;i++){
    idx=seam[i];
    for(j=idx;j<inpic->common.width-inpic->vseamcount;j++)
      rowptr[j]=rowptr[j+1];
    
    rowptr+=inpic->common.width;
  }

  return 1;

}

int removeHorizontalSeam(SEAM_UNWEAVER *inpic, int *seam){
  int i,j;
  int idx;
  int ewidth = inpic->common.width-inpic->vseamcount;
  int eheight=inpic->common.height-inpic->hseamcount;
  int minidx,imin;

  minidx=0;
  for(j=1;j<ewidth;j++){
    if(seam[minidx]>seam[j])
      minidx=j;
  }
  imin=seam[minidx];

  for(i=imin;i<eheight;i++){
    for(j=0;j<ewidth;j++){
      if(i>=seam[j])
	NODE_AT(inpic,i,j)=NODE_AT(inpic,i+1,j);
    }
  }

  return 1;

}


/*It might be a good idea to provide functions to translate seams from the current coordinates to the original coordinates.  The unpicker interface already provides this.*/


/*seam must have room for at least pic->common.height entries.  If translate=1 then the seam returned is in terms of the original image, otherwise the seam returned is in terms of the image with previously identified seams removed.*/
int unweaverNextVertical(SEAM_UNWEAVER *pic, int *seam,int *error){

 if(pic->vseamcount>=pic->common.width){
    SET_ERROR(SEAMSTRESS_E_FINISHED);
    return 0;
  }

  minimalVerticalSeams(pic);
  findMinimalVerticalSeam(pic,seam);
  removeVerticalSeam(pic,seam);
  SET_ERROR(SEAMSTRESS_E_OK);
  return 1;
  
}

/*seam must have room for at least pic->common.height entries.  If translate=1 then the seam returned is in terms of the original image, otherwise the seam returned is in terms of the image with previously identified seams removed.*/
int unweaverNextHorizontal(SEAM_UNWEAVER *pic, int *seam,int *error){
  
  if(pic->hseamcount>=pic->common.height){
    SET_ERROR(SEAMSTRESS_E_FINISHED);
    return 0;
  }

  minimalHorizontalSeams(pic);
  findMinimalHorizontalSeam(pic,seam);
  removeHorizontalSeam(pic,seam);
  SET_ERROR(SEAMSTRESS_E_OK);
  return 1;
  
}


void unweaverCurrentSize(SEAM_UNWEAVER *weaver, int *widthptr,int *heightptr){
  *widthptr=weaver->common.width-weaver->vseamcount;
  *heightptr=weaver->common.height-weaver->hseamcount;
}

/*Converts (*xptr,*yptr) from the current image coordinates to the original image coordinates.*/
int unweaverTransformToOriginal(SEAM_UNWEAVER *weaver, int *xptr, int *yptr){
  unsigned  x,y;

  x=*xptr;
  y=*yptr;

  if((y>=weaver->common.height-weaver->hseamcount)||(x>=weaver->common.width-weaver->vseamcount))
    return 0;

  *xptr=NODE_AT(weaver,y,x).xoffset;
  *yptr=NODE_AT(weaver,y,x).yoffset;
  
  return 1;
}
